"""pyrh models and schemas."""

from .instrument import (
    Instrument,
    InstrumentManager,
    InstrumentPaginator,
    InstrumentPaginatorSchema,
    InstrumentSchema,
)
from .oauth import Challenge, ChallengeSchema, OAuth, OAuthSchema
from .portfolio import Portfolio, PortfolioSchema
from .sessionmanager import SessionManager, SessionManagerSchema

__all__ = [
    "OAuth",
    "OAuthSchema",
    "Challenge",
    "ChallengeSchema",
    "SessionManager",
    "SessionManagerSchema",
    "Portfolio",
    "PortfolioSchema",
    "Instrument",
    "InstrumentSchema",
    "InstrumentManager",
    "InstrumentPaginator",
    "InstrumentPaginatorSchema",
]



"""Manage Robinhood Sessions."""

import uuid
from datetime import datetime, timedelta
from typing import TYPE_CHECKING, Any, Dict, Optional, Union, cast
from urllib.request import getproxies

import certifi
import pyotp as pyotp
import pytz
import requests
from marshmallow import Schema, fields, post_load
from requests.exceptions import HTTPError
from requests.structures import CaseInsensitiveDict
from yarl import URL

from pyrh import urls
from pyrh.exceptions import AuthenticationError, PyrhValueError

from .base import JSON, BaseModel, BaseSchema
from .oauth import CHALLENGE_TYPE_VAL, OAuth, OAuthSchema

# TODO: merge get and post duplicated code into a single function.

# Types
if TYPE_CHECKING:  # pragma: no cover
    CaseInsensitiveDictType = CaseInsensitiveDict[str]
else:
    CaseInsensitiveDictType = CaseInsensitiveDict
Proxies = Dict[str, str]


# Constants
CLIENT_ID: str = "c82SH0WZOsabOXGP2sxqcj34FxkvfnWRZBKlBjFS"
"""Robinhood client id."""
HEADERS: CaseInsensitiveDictType = CaseInsensitiveDict(
    {
        "Accept": "*/*",
        "Accept-Encoding": "gzip, deflate",
        "Accept-Language": "en;q=1, fr;q=0.9, de;q=0.8, ja;q=0.7, nl;q=0.6, it;q=0.5",
        "Content-Type": "application/x-www-form-urlencoded; charset=utf-8",
        "X-Robinhood-API-Version": "1.0.0",
        "Connection": "keep-alive",
        "User-Agent": "Robinhood/823 (iPhone; iOS 7.1.2; Scale/2.00)",
    }
)
"""Headers used when performing requests with robinhood api."""
# 8.5 days (you have a small window to refresh after this)
# I would refresh the token proactively every day in a script
EXPIRATION_TIME: int = 734000
"""Default expiration time for requests."""

TIMEOUT: int = 3
"""Default timeout in seconds"""


class SessionManager(BaseModel):
    """Manage connectivity with Robinhood API.

    This class manages logging in and multi-factor authentication. Optionally,
    it can automatically authenticate for automation systems

    Example:
        >>> sm = SessionManager(username="USERNAME", password="PASSWORD")
        >>> sm.login()  # xdoctest: +SKIP
        >>> sm.logout()  # xdoctest: +SKIP

    Example:
        >>> sm = SessionManager(username="USERNAME", password="PASSWORD", mfa="16DIGITQRCODE")

    If you want to bypass manual MFA authentication, you can supply your 16-digit QR code from Robinhood as a parameter
    as shown.

    Args:
        username: The username to login to Robinhood
        password: The password to login to Robinhood
        mfa: The 16 character QR code used to authenticate MFA automatically
        challenge_type: Either sms or email (only if not using mfa)
        headers: Any optional header dict modifications for the session
        proxies: Any optional proxy dict modification for the session
        **kwargs: Any other passed parameters as converted to instance attributes

    Attributes:
        session: A requests session instance
        expires_at: The time the oauth token will expire at, default is 1970-01-01 00:00:00
        certs: The path to the desired certs to check against
        device_token: A random guid representing the current device
        access_token: An oauth2 token to connect to the Robinhood API
        refresh_token: An oauth2 refresh token to refresh the access_token when required

    """

    def __init__(
        self,
        username: str,
        password: str,
        mfa: Optional[str] = "",
        challenge_type: Optional[str] = "sms",
        headers: Optional[CaseInsensitiveDictType] = None,
        proxies: Optional[Proxies] = None,
        **kwargs: Any,
    ) -> None:
        self.mfa = mfa
        self.session: requests.Session = requests.session()
        self.session.headers = HEADERS if headers is None else headers
        self.session.proxies = getproxies() if proxies is None else proxies
        self.session.verify = certifi.where()
        self.expires_at = datetime.strptime("1970", "%Y").replace(
            tzinfo=pytz.UTC
        )  # some time in the past

        self.username: str = username
        self.password: str = password
        if challenge_type not in ["email", "sms"]:
            raise ValueError("challenge_type must be email or sms")
        self.challenge_type: str = challenge_type

        self.device_token: str = kwargs.pop("device_token", str(uuid.uuid4()))
        self.oauth: OAuth = kwargs.pop("oauth", OAuth())

        super().__init__(**kwargs)

    @property
    def token_expired(self) -> bool:
        """Check if the issued auth token has expired.

        Returns:
            True if expired otherwise False
        """
        return datetime.now(tz=pytz.UTC) > self.expires_at

    @property
    def login_set(self) -> bool:
        """Check if login info is properly configured.

        Returns:
            Whether or not username and password are set.
        """
        return self.password is not None and self.username is not None

    @property
    def authenticated(self) -> bool:
        """Check if the session is authenticated.

        Returns:
            Whether or not the session is logged in.
        """
        return "Authorization" in self.session.headers and not self.token_expired

    def login(self, force_refresh: bool = False) -> None:
        """Login to the session.

        This method logs the user in if they are not already and otherwise refreshes
        the oauth token if it is expired.

        Args:
            force_refresh: If already logged in, whether or not to force a oauth token
                refresh.

        """
        if "Authorization" not in self.session.headers:
            self._login_oauth2()
        elif self.oauth.is_valid and (self.token_expired or force_refresh):
            self._refresh_oauth2()

    def get(
        self,
        url: Union[str, URL],
        params: Optional[Dict[str, Any]] = None,
        headers: Optional[CaseInsensitiveDictType] = None,
        raise_errors: bool = True,
        return_response: bool = False,
        auto_login: bool = True,
        schema: Optional[Schema] = None,
        many: bool = False,
    ) -> Any:
        """Run a wrapped session HTTP GET request.

        Note:
            This method automatically prompts the user to log in if not already logged
            in.

        Args:
            url: The url to get from.
            params: query string parameters
            headers: A dict adding to and overriding the session headers.
            raise_errors: Whether or not raise errors on GET request result.
            return_response: Whether or not return a `requests.Response` object or the
                JSON response from the request.
            auto_login: Whether or not to automatically login on restricted endpoint
                errors.
            schema: An instance of a `marshmallow.Schema` that represents the object
                to build.
            many: Whether to treat the output as a list of the passed schema.

        Returns:
            A JSON dictionary or a constructed object if a schema is passed. If \
                `return_response` is set then a tuple of (response, data) is passed.

        Raises:
            PyrhValueError: If the schema is not an instance of `Schema` and is instead
                a class.

        """
        # Guard against common gotcha, passing schema class instead of instance.
        if isinstance(schema, type):
            raise PyrhValueError("Passed Schema should be an instance not a class.")
        params = {} if params is None else params
        res = self.session.get(
            str(url),
            params=params,
            timeout=TIMEOUT,
            headers={} if headers is None else headers,
        )
        if res.status_code == 401 and auto_login:
            self.login(force_refresh=True)
            res = self.session.get(
                str(url),
                params=params,
                timeout=TIMEOUT,
                headers={} if headers is None else headers,
            )
        if raise_errors:
            res.raise_for_status()

        data = res.json() if schema is None else schema.load(res.json(), many=many)

        return (data, res) if return_response else data

    def post(
        self,
        url: Union[str, URL],
        data: Optional[JSON] = None,
        headers: Optional[CaseInsensitiveDictType] = None,
        raise_errors: bool = True,
        return_response: bool = False,
        auto_login: bool = True,
        schema: Optional[Schema] = None,
        many: bool = False,
    ) -> Any:
        """Run a wrapped session HTTP POST request.

        Note:
            This method automatically prompts the user to log in if not already logged
            in.

        Args:
            url: The url to post to.
            data: The payload to POST to the endpoint.
            headers: A dict adding to and overriding the session headers.
            return_response: Whether or not return a `requests.Response` object or the
                JSON response from the request.
            raise_errors: Whether or not raise errors on POST request.
            auto_login: Whether or not to automatically login on restricted endpoint
                errors.
            schema: An instance of a `marshmallow.Schema` that represents the object
                to build.
            many: Whether to treat the output as a list of the passed schema.

        Returns:
            A JSON dictionary or a constructed object if a schema is passed. If \
                `return_response` is set then a tuple of (response, data) is passed.

        Raises:
            PyrhValueError: If the schema is not an instance of `Schema` and is instead
                a class.

        """
        # Guard against common gotcha, passing schema class instead of instance.
        if isinstance(schema, type):
            raise PyrhValueError("Passed Schema should be an instance not a class.")

        res = self.session.post(
            str(url),
            data=data,
            timeout=TIMEOUT,
            headers={} if headers is None else headers,
        )
        if (res.status_code == 401) and auto_login:
            self.login(force_refresh=True)
            res = self.session.post(
                str(url),
                data=data,
                timeout=TIMEOUT,
                headers={} if headers is None else headers,
            )
        if raise_errors:
            res.raise_for_status()

        data = res.json() if schema is None else schema.load(res.json(), many=many)

        return (data, res) if return_response else data

    def _configure_manager(self, oauth: OAuth) -> None:
        """Process an authentication response dictionary.

        This method updates the internal state of the session based on a login or
        token refresh request.

        Args:
            oauth: An oauth response model from a login request.

        """
        self.oauth = oauth
        self.expires_at = datetime.now(tz=pytz.UTC) + timedelta(
            seconds=self.oauth.expires_in
        )
        self.session.headers.update(
            {"Authorization": f"Bearer {self.oauth.access_token}"}
        )

    def _challenge_oauth2(self, oauth: OAuth, oauth_payload: JSON) -> OAuth:
        """Process the ouath challenge flow.

        Args:
            oauth: An oauth response model from a login request.
            oauth_payload: The payload to use once the challenge has been processed.

        Returns:
            An OAuth response model from the login request.

        Raises:
            AuthenticationError: If there is an error in the initial challenge response.

        .. # noqa: DAR202
        .. https://github.com/terrencepreilly/darglint/issues/81

        """
        # login challenge
        challenge_url = urls.build_challenge(oauth.challenge.id)
        print(
            f"Input challenge code from {oauth.challenge.type.capitalize()} "
            f"({oauth.challenge.remaining_attempts}/"
            f"{oauth.challenge.remaining_retries}):"
        )
        challenge_code = input()
        challenge_payload = {"response": str(challenge_code)}
        challenge_header = CaseInsensitiveDict(
            {"X-ROBINHOOD-CHALLENGE-RESPONSE-ID": str(oauth.challenge.id)}
        )
        oauth_inner, res = self.post(
            challenge_url,
            data=challenge_payload,
            raise_errors=False,
            headers=challenge_header,
            auto_login=False,
            return_response=True,
            schema=OAuthSchema(),
        )
        if res.status_code == requests.codes.ok:
            try:
                # the cast is required for mypy
                return cast(
                    OAuth,
                    self.post(
                        urls.OAUTH,
                        data=oauth_payload,
                        headers=challenge_header,
                        auto_login=False,
                        schema=OAuthSchema(),
                    ),
                )
            except HTTPError:
                raise AuthenticationError("Error in finalizing auth token")
        elif oauth_inner.is_challenge and oauth_inner.challenge.can_retry:
            print("Invalid code entered")
            return self._challenge_oauth2(oauth, oauth_payload)
        else:
            raise AuthenticationError("Exceeded available attempts or code expired")

    def _mfa_oauth2(self, oauth_payload: JSON, attempts: int = 3) -> OAuth:
        """Mfa auth flow.

         For people with 2fa.

        Args:
            oauth_payload: JSON payload to send on mfa approval.
            attempts: The number of attempts to allow for mfa approval.

        Returns:
            An OAuth response model object.

        Raises:
            AuthenticationError: If the mfa code is incorrect more than specified \
                number of attempts.

        """
        print("Input mfa code:")
        if self.mfa != "":
            mfa_code = pyotp.TOTP(self.mfa).now()
        else:
            mfa_code = input()
        oauth_payload["mfa_code"] = mfa_code
        oauth, res = self.post(
            urls.OAUTH,
            data=oauth_payload,
            raise_errors=False,
            auto_login=False,
            return_response=True,
            schema=OAuthSchema(),
        )
        attempts -= 1
        if (res.status_code != requests.codes.ok) and (attempts > 0):
            print("Invalid mfa code")
            return self._mfa_oauth2(oauth_payload, attempts)
        elif res.status_code == requests.codes.ok:
            # TODO: Write mypy issue on why this needs to be casted?
            return cast(OAuth, oauth)
        else:
            raise AuthenticationError("Too many incorrect mfa attempts")

    def _login_oauth2(self) -> None:
        """Create a new oauth2 token.

        Raises:
            AuthenticationError: If the login credentials are not set, if a challenge
                wasn't accepted, or if an mfa code is not accepted.

        """
        self.session.headers.pop("Authorization", None)

        oauth_payload = {
            "password": self.password,
            "username": self.username,
            "grant_type": "password",
            "client_id": CLIENT_ID,
            "expires_in": EXPIRATION_TIME,
            "scope": "internal",
            "device_token": self.device_token,
            "challenge_type": self.challenge_type,
        }

        oauth = self.post(
            urls.OAUTH,
            data=oauth_payload,
            raise_errors=False,
            auto_login=False,
            schema=OAuthSchema(),
        )

        if oauth.is_challenge:
            oauth = self._challenge_oauth2(oauth, oauth_payload)
        elif oauth.is_mfa:
            oauth = self._mfa_oauth2(oauth_payload)

        if not oauth.is_valid:
            if hasattr(oauth, "error"):
                msg = f"{oauth.error}"
            elif hasattr(oauth, "detail"):
                msg = f"{oauth.detail}"
            else:
                msg = "Unknown login error"
            raise AuthenticationError(msg)
        else:
            self._configure_manager(oauth)

    def _refresh_oauth2(self) -> None:
        """Refresh an oauth2 token.

        Raises:
            AuthenticationError: If refresh_token is missing or if there is an error
                when trying to refresh a token.

        """
        if not self.oauth.is_valid:
            raise AuthenticationError("Cannot refresh login with unset refresh token")
        relogin_payload = {
            "grant_type": "refresh_token",
            "refresh_token": self.oauth.refresh_token,
            "scope": "internal",
            "client_id": CLIENT_ID,
            "expires_in": EXPIRATION_TIME,
        }
        self.session.headers.pop("Authorization", None)
        try:
            oauth = self.post(
                urls.OAUTH,
                data=relogin_payload,
                auto_login=False,
                schema=OAuthSchema(),
            )
        except HTTPError:
            raise AuthenticationError("Failed to refresh token")

        self._configure_manager(oauth)

    def logout(self) -> None:
        """Logout from the session.

        Raises:
            AuthenticationError: If there is an error when logging out.

        """
        logout_payload = {"client_id": CLIENT_ID, "token": self.oauth.refresh_token}
        try:
            self.post(urls.OAUTH_REVOKE, data=logout_payload, auto_login=False)
            self.oauth = OAuth()
            self.session.headers.pop("Authorization", None)
        except HTTPError:
            raise AuthenticationError("Could not log out")

    def __repr__(self) -> str:
        """Return the object as a string.

        Returns:
            The string representation of the object.

        """
        return f"SessionManager<{self.username}>"


class SessionManagerSchema(BaseSchema):
    """Schema class for the SessionManager model."""

    __model__ = SessionManager

    # Call untyped "Email" in typed context
    username = fields.Email()  # type: ignore
    password = fields.Str()
    challenge_type = fields.Str(validate=CHALLENGE_TYPE_VAL)
    oauth = fields.Nested(OAuthSchema)
    expires_at = fields.AwareDateTime()
    device_token = fields.Str()
    headers = fields.Dict()
    proxies = fields.Dict()

    @post_load
    def make_object(self, data: JSON, **kwargs: Any) -> SessionManager:
        """Override default method to configure SessionManager object on load.

        Args:
            data: The JSON dictionary to process
            **kwargs: Not used but matches signature of `BaseSchema.make_object`

        Returns:
            A configured instance of SessionManager.
        """
        oauth = data.pop("oauth", None)
        expires_at = data.pop("expires_at", None)
        session_manager = self.__model__(**data)

        if oauth is not None and oauth.is_valid:
            session_manager.oauth = oauth
            session_manager.session.headers.update(
                {"Authorization": f"Bearer {session_manager.oauth.access_token}"}
            )
        if expires_at:
            session_manager.expires_at = expires_at

        return session_manager



"""Define Robinhood endpoints."""

from typing import Optional

from yarl import URL

# TODO: All url construction should happen here, not in robinhood.py

# Base
API_BASE = URL("https://api.robinhood.com")

# General
ACCOUNTS = API_BASE / "accounts/"
ACH_BASE = API_BASE / "ach/"  # not implemented
APPLICATIONS = API_BASE / "applications/"  # not implemented
DIVIDENDS = API_BASE / "dividends/"
DOCUMENTS = API_BASE / "documents/"  # not implemented
DOCUMENT_REQUESTS = API_BASE / "upload/document_requests/"  # not implemented
FUNDAMENTALS_BASE = API_BASE / "fundamentals/"
INSTRUMENTS_BASE = API_BASE / "instruments/"
MARGIN_UPGRADES = API_BASE / "margin/upgrades/"  # not implemented
MARKETS = API_BASE / "markets/"  # not implemented
MARKET_DATA_BASE = API_BASE / "marketdata/"
NEWS_BASE = API_BASE / "midlands/news/"
NOTIFICATIONS = API_BASE / "notifications/"  # not implemented
ORDERS_BASE = API_BASE / "orders/"
PORTFOLIOS = API_BASE / "portfolios/"
POSITIONS = API_BASE / "positions/"
TAGS_BASE = API_BASE / "midlands/tags/tag/"
WATCHLISTS = API_BASE / "watchlists/"

# Options
OPTIONS_BASE = API_BASE / "options/"
OPTIONS_CHAIN_BASE = OPTIONS_BASE / "chains/"
OPTIONS_INSTRUMENTS_BASE = OPTIONS_BASE / "instruments/"

# User
USER = API_BASE / "user/"
INVESTMENT_PROFILE = USER / "investment_profile/"

# Quotes
QUOTES = API_BASE / "quotes/"
HISTORICALS = QUOTES / "historicals/"

# Auth
OAUTH_BASE: URL = API_BASE / "oauth2/"
OAUTH: URL = OAUTH_BASE / "token/"
OAUTH_REVOKE: URL = OAUTH_BASE / "revoke_token/"
MIGRATE_TOKEN: URL = OAUTH_BASE / "migrate_token/"  # not implemented
PASSWORD_RESET: URL = API_BASE / "password_reset/request/"  # not implemented


def build_challenge(challenge_id: str) -> URL:
    """Build challenge response url.

    Args:
        challenge_id: the id of the challenge passed in the oauth request flow.

    Returns:
        The constructed URL with the challenge_id embedded in teh url path.

    """
    return API_BASE / f"challenge/{challenge_id}/respond/"


def build_ach(option: str) -> URL:
    """
    Combination of 3 ACH endpoints. Options include:
        * iav
        * relationships
        * transfers
    """
    return ACH_BASE / "iav/auth/" if option == "iav" else ACH_BASE / f"{option}/"


def instruments(
    symbol: Optional[str] = None, query: Optional[str] = None, id_: Optional[str] = None
) -> URL:
    """Construct urls that query instruments.

    Note:
        Each of the arguments are mutually exclusive.

    Args:
        symbol: A stock ticker symbol.
        query: Keyword to search for an instrument. (might be in name or ticker)
        id_: The UUID that represents the instrument.

    Returns:
        A constructed URL with the embedded query parameter

    """
    # Note:
    # INSTRUMENTS_BASE/{instrument_id}/splits will not be implemented since the url is
    # embedded in the results of an individual instrument result. The same logic applies
    # for INSTRUMENTS_BASE/{instrument_id}/splits/{split_id}/
    if symbol is not None:
        return INSTRUMENTS_BASE.with_query(symbol=symbol)
    elif query is not None:
        return INSTRUMENTS_BASE.with_query(query=query)
    elif id_ is not None:
        return INSTRUMENTS_BASE / f"{id_}/"


def build_orders(order_id: Optional[str] = None) -> URL:
    """Build endpoint to place orders."

    Args:
        order_id: the id of the order

    Returns:
        A constructed URL for a particular order or the base URL for all orders.

    """
    if order_id is not None:
        return ORDERS_BASE / f"{order_id}/"
    else:
        return ORDERS_BASE


def build_news(stock: str) -> URL:
    """Build news endpoint for a particular stock

    Args:
        stock: The stock ticker to build the URL

    Returns:
        A constructed URL for the input stock ticker.

    """
    return NEWS_BASE / f"{stock}/"


def build_fundamentals(stock: str) -> URL:
    """Build fundamentals endpoint for a particular stock

    Args:
        stock: The stock ticker to build the URL

    Returns:
        A constructed URL of the fundamentals for the input stock ticker.

    """
    return FUNDAMENTALS_BASE / f"{stock}/"


def build_tags(tag: str) -> URL:
    """Build endpoints for tickers with a particular tag.

    Args:
        tag: The tag to search for.

    Returns:
        A constructed URL for a particular tag.

    """
    return TAGS_BASE / f"{tag}/"


def build_chain(instrument_id: str) -> URL:
    """Build the query for a particular options chain.

    # TODO: this isn't best practice
    # (query construction should be a separate function)

    Args:
        instrument_id: The instrument in question.

    Returns:
        A constructed URL for the particular options chain search.

    """
    return (
        OPTIONS_CHAIN_BASE.with_query(equity_instrument_ids=f"{instrument_id}") / "/"
    )  # TODO: find out if this trailing slash is required.


def build_options(chain_id: str, dates: str, option_type: str) -> URL:
    """Build options search endpoint.

    # TODO: this really isn't best practice.

    Args:
        chain_id: The id for a particular options chain.
        dates: The range of dates to procure # TODO: document the format of the dates
        option_type: The type of the option # TODO: document the types
    """
    return OPTIONS_INSTRUMENTS_BASE.with_query(
        chain_id=f"{chain_id}",
        expiration_dates=f"{dates}",
        state="active",
        tradability="tradable",
        type=f"{option_type}",
    )


def build_market_data(option_id: Optional[str] = None) -> URL:
    """Build market data endpoint.

    Args:
        option_id: the id of the option.

    Returns:
        A constructed URL for market data for a particular `option_id`.
    """
    if option_id is None:
        return MARKET_DATA_BASE / f"{option_id}/"
    else:
        return MARKET_DATA_BASE



"""Oauth models."""

from datetime import datetime

import pytz
from marshmallow import fields, validate

from .base import BaseModel, BaseSchema

CHALLENGE_TYPE_VAL = validate.OneOf(["email", "sms"])


class Challenge(BaseModel):
    """The challenge response model."""

    remaining_attempts = 0
    """Default `remaining_attempts` attribute if it is not set on instance."""

    @property
    def can_retry(self) -> bool:
        """Determine if the challenge can be retried.

        Returns:
            True if remaining_attempts is greater than zero and challenge is not \
                expired, False otherwise.

        """
        return self.remaining_attempts > 0 and (
            datetime.now(tz=pytz.utc) < self.expires_at
        )


class ChallengeSchema(BaseSchema):
    """The challenge response schema."""

    __model__ = Challenge

    id = fields.UUID()
    user = fields.UUID()
    type = fields.Str(validate=CHALLENGE_TYPE_VAL)
    alternate_type = fields.Str(
        validate=CHALLENGE_TYPE_VAL, required=False, allow_none=True
    )
    status = fields.Str(validate=validate.OneOf(["issued", "validated", "failed"]))
    remaining_retries = fields.Int()
    remaining_attempts = fields.Int()
    expires_at = fields.AwareDateTime(default_timezone=pytz.UTC)  # type: ignore


class OAuth(BaseModel):
    """The OAuth response model."""

    @property
    def is_challenge(self) -> bool:
        """Determine whether the oauth response is a challenge.

        Returns:
            True response has the `challenge` key, False otherwise.

        """
        return hasattr(self, "challenge")

    @property
    def is_mfa(self) -> bool:
        """Determine whether the oauth response is a mfa challenge.

        Returns:
            True response has the `mfa_required` key, False otherwise.

        """
        return hasattr(self, "mfa_required")

    @property
    def is_valid(self) -> bool:
        """Determine whether the oauth response is a valid response.

        Returns:
            True if the response has both the `access_token` and `refresh_token` keys, \
                False otherwise.

        """
        return hasattr(self, "access_token") and hasattr(self, "refresh_token")


class OAuthSchema(BaseSchema):
    """The OAuth response schema."""

    __model__ = OAuth

    detail = fields.Str()
    challenge = fields.Nested(ChallengeSchema)
    mfa_required = fields.Boolean()

    access_token = fields.Str()
    refresh_token = fields.Str()
    expires_in = fields.Int()



"""Base Model."""
from collections.abc import MutableSequence
from types import SimpleNamespace
from typing import Any, Callable, Dict, Iterable, Mapping, Optional

from marshmallow import INCLUDE, Schema, fields, post_load
from yarl import URL

from pyrh.exceptions import InvalidOperation

JSON = Dict[str, Any]
MAX_REPR_LEN = 50


def _process_dict_values(value: Any) -> Any:
    """Process a returned from a JSON response.

    Args:
        value: A dict, list, or value returned from a JSON response.

    Returns:
        Either an UnknownModel, a List of processed values, or the original value \
            passed through.

    """
    if isinstance(value, Mapping):
        return UnknownModel(**value)
    elif isinstance(value, list):
        return [_process_dict_values(v) for v in value]
    else:
        return value


class BaseModel(SimpleNamespace):
    """BaseModel that all models should inherit from.

    Note:
        If a passed parameter is a nested dictionary, then it is created with the
        `UnknownModel` class. If it is a list, then it is created with

    Args:
        **kwargs: All passed parameters as converted to instance attributes.
    """

    def __init__(self, **kwargs: Any) -> None:
        kwargs = {k: _process_dict_values(v) for k, v in kwargs.items()}

        self.__dict__.update(kwargs)

    def __repr__(self) -> str:
        """Return a default repr of any Model.

        Returns:
            The string model parameters up to a `MAX_REPR_LEN`.

        """
        repr_ = super().__repr__()
        if len(repr_) > MAX_REPR_LEN:
            return repr_[:MAX_REPR_LEN] + " ...)"
        else:
            return repr_


class UnknownModel(BaseModel):
    """A convenience class that inherits from `BaseModel`."""

    pass


class BaseSchema(Schema):
    """The default schema for all models."""

    __model__: Any = UnknownModel
    """Determine the object that is created when the load method is called."""
    __first__: Optional[str] = None
    """Determine if `make_object` will try to get the first element the input key."""

    class Meta:
        unknown = INCLUDE

    @post_load
    def make_object(self, data: JSON, **kwargs: Any) -> "__model__":
        """Build model for the given `__model__` class attribute.

        Args:
            data: The JSON diction to use to build the model.
            **kwargs: Unused but required to match signature of `Schema.make_object`

        Returns:
            An instance of the `__model__` class.

        """
        if self.__first__ is not None:
            data_list = data.get("results", [{}])
            # guard against empty return list of a valid results return
            data = data_list[0] if len(data_list) != 0 else {}
        return self.__model__(**data)


def has_results(func: Callable[..., Any]) -> Callable[..., Any]:
    """Check whether a particular function has results when filtering its data.

    Args:
        func: The function to be decorated

    Returns:
        The decorated function which raises `InvalidOperation` if the \
            decorated function does not have the `results` attribute

    """

    def _decorator(self: Any, *args: Any, **kwargs: Any) -> Any:
        if not hasattr(self, "results") or self.results is None:
            raise InvalidOperation(
                "The result attribute cannot be None for this method call."
            )
        return func(self, *args, **kwargs)

    return _decorator


# TODO: figure mypy complains on this line
class BasePaginator(BaseModel, MutableSequence):  # type: ignore
    """Thin wrapper around `self.results` for a robinhood paginator."""

    @has_results
    def __getitem__(self, key: Any) -> Any:  # noqa: D105
        return self.results[key]

    @has_results
    def __setitem__(self, key: Any, value: Any) -> Any:  # noqa: D105
        self.results[key] = value

    @has_results
    def __delitem__(self, key: Any) -> None:  # noqa: D105
        del self.results[key]

    @has_results
    def __len__(self) -> int:  # noqa: D105
        return len(self.results)

    @has_results
    def insert(self, index: int, element: Any) -> None:  # noqa: D102
        self.results.insert(index, element)


class BasePaginatorSchema(BaseSchema):
    """BasePaginatorSchema for the BasePaginator class.

    Note:
        Make sure to re-define the results attribute based on the subclass schema.

    """

    __model__ = BasePaginator

    next = fields.URL(allow_none=True)
    previous = fields.URL(allow_none=True)
    results = fields.List(fields.Nested(UnknownModel))


# TODO: Figure how to resolve the circular import with SessionManager (type ignore)
def base_paginator(
    seed_url: "URL", session_manager: Any, schema: Any
) -> Iterable[Any]:  # type: ignore  # noqa: F821
    """Create a paginator using the passed parameters.

    Args:
        seed_url: The url to get the first batch of results.
        session_manager: The session manager that will manage the get.
        schema: The Schema subclass used to build individual instances.

    Yields:
        Instances of the object passed in the schema field.

    """
    resource_endpoint = seed_url
    while True:
        paginator = session_manager.get(resource_endpoint, schema=schema)
        for instrument in paginator:
            yield instrument
        if paginator.next is not None:
            resource_endpoint = paginator.next
        else:
            break



"""Stock Instruments in Robinhood."""

from typing import Iterable, Optional, cast

from marshmallow import fields

from pyrh import urls
from pyrh.exceptions import PyrhValueError

from .base import (
    BaseModel,
    BasePaginator,
    BasePaginatorSchema,
    BaseSchema,
    base_paginator,
)
from .sessionmanager import SessionManager


# TODO: dream up a good way to not require session to `get_fundamentals` without a
# Singleton pattern since there could be multiple sessions in the future.
class Instrument(BaseModel):
    """A financial instrument."""

    def get_fundamentals(self):  # type: ignore
        """TODO."""  # noqa: DAR401
        raise NotImplementedError()

    def get_market(self):  # type: ignore
        """TODO."""  # noqa: DAR401
        raise NotImplementedError()

    def get_quote(self):  # type: ignore
        """TODO."""  # noqa: DAR401
        raise NotImplementedError()

    def get_splits(self):  # type: ignore
        """TODO."""  # noqa: DAR401
        raise NotImplementedError()


class InstrumentSchema(BaseSchema):
    """The Schema for Instrument objects."""

    __model__ = Instrument
    # symbol query returns a paginator but we only want the first
    __first__ = "results"

    bloomberg_unique = fields.Str()
    country = fields.Str()
    day_trade_ratio = fields.Float()
    default_collar_fraction = fields.Float()
    fractional_tradability = fields.Str()  # TODO: determine possible values
    fundamentals = fields.URL()
    id = fields.UUID()
    list_date = fields.NaiveDateTime(format="%Y-%m-%d", allow_none=True)
    maintenance_ratio = fields.Float()
    margin_initial_ratio = fields.Float()
    market = fields.URL()
    # This value can be null: http://www.finra.org/industry/tick-size-pilot-program
    min_tick_size = fields.Float(allow_none=True)
    name = fields.Str()
    quote = fields.URL()
    rhs_tradability = fields.Str()  # TODO: determine possible values
    simple_name = fields.Str(allow_none=True)
    splits = fields.URL()
    state = fields.Str()  # TODO: determine possible values
    symbol = fields.Str()
    tradability = fields.Str()
    tradable_chain_id = fields.Str(allow_none=True)  # TODO: determine possible values
    tradeable = fields.Boolean()  # looks like they're mixing UK and US english
    type = fields.Str()  # TODO: determine possible values
    url = fields.URL()


class InstrumentPaginator(BasePaginator):
    """Thin wrapper around `self.results`, a list of `Instruments`."""

    pass


class InstrumentPaginatorSchema(BasePaginatorSchema):
    """Schema class for the InstrumentPaginator.

    The nested results are of types `Instrument`.

    """

    __model__ = InstrumentPaginator

    results = fields.List(fields.Nested(InstrumentSchema))


class InstrumentManager(SessionManager):
    """Group together methods that manipulate instruments.

    Examples:
        >>> im = InstrumentManager()
        >>> im.instruments()  # Get all instruments
        >>> im.instrument(symbol="TSLA")  # Get a particular instrument

    """

    def instruments(self, query: Optional[str] = None) -> Iterable[Instrument]:
        """Get a generator of instruments.

        Args:
            query: If the query argument is provided, the returned values will be
                restricted to instruments that match the query keyword (single word)

        Returns:
            A generator of Instruments.

        """
        url = (
            urls.INSTRUMENTS_BASE
            if query is not None
            else urls.instruments(query=query)
        )
        return base_paginator(url, self, InstrumentPaginatorSchema())

    def instrument(
        self, symbol: Optional[str] = None, id_: Optional[str] = None
    ) -> Instrument:
        """Get a single instrument using a provided query parameter.

        Note:
            The input parameters are mutually exclusive. Additionally, if you query a
            hidden symbol it will return emtpy. The only way to view hidden symbols is
            to use the instruments endpoint.

        Args:
            symbol: A ticker symbol
            id_: A UUID that represents an instrument

        Returns:
            A single instance of an `Instrument`

        Raises:
            PyrhValueError: Neither of the input kwargs are passed in.

        """
        if any(opt is not None for opt in [symbol, id_]):
            return cast(
                Instrument,
                self.get(
                    urls.instruments(symbol=symbol, id_=id_), schema=InstrumentSchema()
                ),
            )
        else:
            raise PyrhValueError("No valid options were provided.")



"""Exceptions: custom exceptions for library"""


class PyrhException(Exception):
    """Wrapper for custom robinhood library exceptions."""

    pass


class PyrhValueError(ValueError, PyrhException):
    """Value Error for the pyrh library."""


class InvalidCacheFile(PyrhException):
    """Error when the cache config file is found to be invalid."""

    pass


class InvalidOperation(PyrhException):
    """An invalid operation was requsted to be performed."""

    pass


class AuthenticationError(PyrhException):
    """Error when trying to login to robinhood."""

    pass


class InvalidTickerSymbol(PyrhException):
    """When an invalid ticker (stock symbol) is given/"""

    pass


class InvalidOptionId(PyrhException):
    """When an invalid option id is given/"""

    pass



"""Current portfolio."""

from marshmallow import fields

from .base import BaseModel, BaseSchema


class Portfolio(BaseModel):
    """Robinhood portfolio data class."""

    pass


class PortfolioSchema(BaseSchema):
    """Robinhood portfolio schema data loader."""

    __model__ = Portfolio
    __first__ = "results"  # returns a paginator but we only need the first

    url = fields.URL()
    account = fields.URL()
    start_date = fields.NaiveDateTime()
    market_value = fields.Float()
    equity = fields.Float()
    extended_hours_market_value = fields.Float()
    extended_hours_equity = fields.Float()
    extended_hours_portfolio_equity = fields.Float()
    last_core_market_value = fields.Float()
    last_core_equity = fields.Float()
    last_core_portfolio_equity = fields.Float()
    excess_margin = fields.Float()
    excess_maintenance = fields.Float()
    excess_margin_with_uncleared_deposits = fields.Float()
    portfolio_equity_previous_close = fields.Float()
    adjusted_equity_previous_close = fields.Float()
    adjusted_portfolio_equity_previous_close = fields.Float()
    withdrawable_amount = fields.Float()
    unwithdrawable_deposits = fields.Float()
    unwithdrawable_grants = fields.Float()



"""Export pyrh sub classes."""

from . import exceptions
from .cache import dump_session, load_session
from .robinhood import Robinhood

__version__ = "2.1.1"
__all__ = [
    "__version__",
    "Robinhood",
    "load_session",
    "dump_session",
    "exceptions",
]



# type: ignore

import csv
import shelve

from pyrh import Robinhood


def get_symbol_from_instrument_url(rb_client, url, db):
    instrument = {}
    if url in db:
        instrument = db[url]
    else:
        db[url] = fetch_json_by_url(rb_client, url)
        instrument = db[url]
    return instrument["symbol"]


def fetch_json_by_url(rb_client, url):
    return rb_client.session.get(url).json()


def order_item_info(order, rb_client, db):
    # side: .side,  price: .average_price, shares: .cumulative_quantity,
    # instrument: .instrument, date : .last_transaction_at
    symbol = get_symbol_from_instrument_url(rb_client, order["instrument"], db)
    return {
        "side": order["side"],
        "price": order["average_price"],
        "shares": order["cumulative_quantity"],
        "symbol": symbol,
        "date": order["last_transaction_at"],
        "state": order["state"],
    }


def get_all_history_orders(rb_client):
    orders = []
    past_orders = rb_client.order_history()
    orders.extend(past_orders["results"])
    while past_orders["next"]:
        print("{} order fetched".format(len(orders)))
        next_url = past_orders["next"]
        past_orders = fetch_json_by_url(rb_client, next_url)
        orders.extend(past_orders["results"])
    print("{} order fetched".format(len(orders)))
    return orders


rb = Robinhood()
# !!!!!! change the username and passs, be careful when paste the code to public
rb.login(username="name", password="pass")
past_orders = get_all_history_orders(rb)
instruments_db = shelve.open("instruments.db")
orders = [order_item_info(order, rb, instruments_db) for order in past_orders]
keys = ["side", "symbol", "shares", "price", "date", "state"]
with open("orders.csv", "w") as output_file:
    dict_writer = csv.DictWriter(output_file, keys)
    dict_writer.writeheader()
    dict_writer.writerows(orders)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pyrh', 'pyrh.models']

package_data = \
{'': ['*']}

install_requires = \
['certifi>=2020.4.5,<2021.0.0',
 'marshmallow>=3.5.1,<4.0.0',
 'pyotp>=2.7.0,<3.0.0',
 'python-dateutil>=2.8,<3.0',
 'pytz>=2019.3,<2020.0',
 'requests>=2.23,<3.0',
 'yarl>=1.4.2,<2.0.0']

extras_require = \
{'docs': ['autodocsumm>=0.2.9,<0.3.0',
          'sphinx>=5.3.0,<6.0.0',
          'sphinx-autodoc-typehints>=1.19.5,<2.0.0',
          'sphinx_rtd_theme>=1.1.1,<2.0.0'],
 'notebook': ['notebook>=6.0.3,<7.0.0', 'python-dotenv>=0.13.0,<0.14.0']}

setup_kwargs = {
    'name': 'pyrh',
    'version': '2.1.1',
    'description': 'Unofficial Robinhood Python API',
    'long_description': '.. image:: https://i.imgur.com/74CYw5g.png\n   :target: https://github.com/robinhood-unofficial/pyrh\n   :alt: robinhood-logo\n\n-------------------------------------------------------------\n\npyrh - Unofficial Robinhood API\n###############################\n\n.. image:: https://github.com/robinhood-unofficial/pyrh/workflows/build/badge.svg?branch=master&event=push\n   :target: https://github.com/robinhood-unofficial/pyrh/actions?query=workflow%3Abuild+branch%3Amaster\n   :alt: Build Status\n\n.. image:: https://codecov.io/gh/robinhood-unofficial/pyrh/branch/master/graph/badge.svg\n   :target: https://codecov.io/gh/robinhood-unofficial/pyrh\n   :alt: Coverage\n\n.. image:: https://readthedocs.org/projects/pyrh/badge/?version=latest\n   :target: https://pyrh.readthedocs.io/en/latest/?badge=latest\n   :alt: Documentation Status\n\n.. image:: https://img.shields.io/pypi/v/pyrh?style=plastic\n   :target: https://pypi.org/project/pyrh/\n   :alt: PyPI Version\n\n.. image:: https://img.shields.io/pypi/dm/pyrh?color=blue&style=plastic\n   :target: https://pypi.org/project/pyrh/\n   :alt: PyPI - Downloads\n\n.. image:: https://img.shields.io/github/license/robinhood-unofficial/Robinhood\n   :target: https://github.com/robinhood-unofficial/pyrh/blob/master/LICENSE\n   :alt: License\n\n.. image:: https://img.shields.io/badge/code%20style-black-000000.svg\n   :target: https://github.com/psf/black\n   :alt: Code Style\n\n.. image:: https://img.shields.io/gitter/room/J-Robinhood/Lobby\n   :target: https://gitter.im/J-Robinhood/Lobby?utm_source=badge&utm_medium=badge&utm_campaign=pr-badge&utm_content=badge\n   :alt: Gitter\n\nPython Framework to make trades with Unofficial Robinhood API. Supports Python 3.8.1+\n\n*Please note this project is not currently actively maintained, but is accepting contributions*\n\nDocumentation: https://pyrh.readthedocs.io/en/latest/\n\nQuick start\n***********\n\n.. code-block:: python\n\n   from pyrh import Robinhood\n\n   rh = Robinhood(username="YOUR_EMAIL", password="YOUR_PASSWORD")\n   rh.login()\n   rh.print_quote("AAPL")\n\nHow To Install:\n***************\n\n.. code-block::\n\n   pip install pyrh\n\nRunning example.ipynb_\n**********************\n\n.. _example.ipynb: https://github.com/robinhood-unofficial/pyrh/blob/master/notebooks/example.ipynb\n\nClone the repository and install jupyter capabilities.\n\n.. code-block::\n\n   $ git clone https://github.com/robinhood-unofficial/pyrh.git\n   $ cd pyrh\n   $ python --version # python 3.3+ for venv functionality\n   Python 3.8.1\n   $ python -m venv pyrh_env\n   $ source pyrh_env/bin/activate\n   (pyrh_env) $ pip install .[notebook]\n   (pyrh_env) $ cp .env.sample .env # update the values in here\n   (pyrh_env) $ jupyter notebook notebooks/example.ipynb\n\nNow just run the files in the example.\n\nRelated\n*******\n\n* `robinhood-ruby <https://github.com/rememberlenny/robinhood-ruby>`_ - RubyGem for interacting with Robinhood API\n* `robinhood-node <https://github.com/aurbano/robinhood-node>`_ - NodeJS module to make trades with Robinhood Private API\n* See the original `blog post <https://medium.com/@rohanpai25/reversing-robinhood-free-accessible-automated-stock-trading-f40fba1e7d8b>`_.\n',
    'author': 'Unofficial Robinhood Python API Developers',
    'author_email': 'None',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://pyrh.readthedocs.io/en/latest',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.8.1,<4.0.0',
}


setup(**setup_kwargs)



"""Project config/cache files."""

from json import JSONDecodeError
from pathlib import Path
from typing import Optional, Union, cast

from pyrh.exceptions import InvalidCacheFile
from pyrh.robinhood import Robinhood, RobinhoodSchema

CACHE_ROOT: Path = Path("~/.robinhood").expanduser()
"""The root directory where cache and config files are stores.

Creates the directory on import.
"""
CACHE_ROOT.mkdir(parents=True, exist_ok=True)

CACHE_LOGIN: Path = CACHE_ROOT.joinpath("login.json")
"""Path to login.json config file.

Creates the file on import.
"""
CACHE_LOGIN.touch(exist_ok=True)

# TODO: Fix darglint issue (remove from flake8 ignore)
# https://github.com/terrencepreilly/darglint/issues/81


def dump_session(robinhood: Robinhood, path: Optional[Union[Path, str]] = None) -> None:
    """Save the current session parameters to a json file.

    Note:
        This function defaults to caching this information to
        ~/.robinhood/login.json

    Args:
        robinhood: A Robinhood instance.
        path: The location to save the file and its name.

    """
    path = CACHE_LOGIN if path is None else path
    json_str = RobinhoodSchema().dumps(robinhood, indent=4)

    with open(path, "w+") as file:
        file.write(json_str)


def load_session(path: Optional[Union[Path, str]] = None) -> Robinhood:
    """Load cached session parameters from a json file.

    Note:
        This function defaults to caching this information to
        ~/.robinhood/login.json

    Args:
        path: The location and file name to load from.

    Returns:
        A configured instance of SessionManager.

    Raises:
        InvalidCacheFile: If the cache file cannot be decoded or does not exist

    """
    path = path or CACHE_LOGIN
    try:
        with open(path) as file:
            return cast(Robinhood, RobinhoodSchema().loads(file.read()))
    except (JSONDecodeError, FileNotFoundError):
        raise InvalidCacheFile(
            f"The cache file at {path} is invalid or does not exist."
        )



"""robinhood.py: a collection of utilities for working with Robinhood's Private API."""

from enum import Enum
from urllib.parse import unquote

import dateutil
import requests
from yarl import URL

from pyrh import urls
from pyrh.exceptions import InvalidTickerSymbol
from pyrh.models import (
    InstrumentManager,
    PortfolioSchema,
    SessionManager,
    SessionManagerSchema,
)

# TODO: re-enable InvalidOptionId when broken endpoint function below is fixed


class Bounds(Enum):
    """Enum for bounds in `historicals` endpoint."""

    REGULAR = "regular"
    EXTENDED = "extended"


class Transaction(Enum):
    """Enum for buy/sell orders."""

    BUY = "buy"
    SELL = "sell"


class Robinhood(InstrumentManager, SessionManager):
    """Wrapper class for fetching/parsing Robinhood endpoints.

    Please see :py:class:`pyrh.models.sessionmanager.SessionManager` for login functionality.

    Provides a global convenience wrapper for the following manager objects:

        * InstrumentManager
        * TODO: Add to this list

    """

    ###########################################################################
    #                               GET DATA                                  #
    ###########################################################################

    def user(self):
        return self.get(urls.USER)

    def investment_profile(self):
        """Fetch investment_profile."""
        return self.get(urls.INVESTMENT_PROFILE)

    def quote_data(self, stock=""):
        """Fetch stock quote.

        Args:
            stock (str or dict): stock ticker symbol or stock instrument

        Returns:
            (:obj:`dict`): JSON contents from `quotes` endpoint

        """

        if isinstance(stock, dict):
            if "symbol" in stock.keys():
                url = str(urls.QUOTES) + stock["symbol"] + "/"
        elif isinstance(stock, str):
            url = str(urls.QUOTES) + stock + "/"
        else:
            raise InvalidTickerSymbol()

        # Check for validity of symbol
        try:
            data = self.get(url)
        except requests.exceptions.HTTPError:
            raise InvalidTickerSymbol()

        return data

    # We will keep for compatibility until next major release
    def quotes_data(self, stocks):
        """Fetch quote for multiple stocks, in one single Robinhood API call.

        Args:
            stocks (list<str>): stock tickers

        Returns:
            (:obj:`list` of :obj:`dict`): List of JSON contents from `quotes` \
                endpoint, in the same order of input args. If any ticker is \
                invalid, a None will occur at that position.

        """

        url = str(urls.QUOTES) + "?symbols=" + ",".join(stocks)

        try:
            data = self.get(url)
        except requests.exceptions.HTTPError:
            raise InvalidTickerSymbol()

        return data["results"]

    def get_quote_list(self, stock="", key=""):
        """Returns multiple stock info and keys from quote_data (prompt if blank)

        Args:
            stock (str): stock ticker (or tickers separated by a comma)
            , prompt if blank
            key (str): key attributes that the function should return

        Returns:
            (:obj:`list`): Returns values from each stock or empty list
                           if none of the stocks were valid

        """

        # Creates a tuple containing the information we want to retrieve
        def append_stock(stock):
            keys = key.split(",")
            myStr = ""
            for item in keys:
                myStr += f"{stock[item]},"

            return myStr.split(",")

        # Prompt for stock if not entered
        if not stock:  # pragma: no cover
            stock = input("Symbol: ")

        data = self.quote_data(stock)
        res = []

        # Handles the case of multple tickers
        if stock.find(",") != -1:
            for stock in data["results"]:
                if stock is None:
                    continue
                res.append(append_stock(stock))

        else:
            res.append(append_stock(data))

        return res

    def get_quote(self, stock=""):
        """Wrapper for quote_data."""

        data = self.quote_data(stock)
        return data

    def get_stock_marketdata(self, instruments):
        info = self.get_url(
            urls.build_market_data() + "quotes/?instruments=" + ",".join(instruments)
        )
        return info["results"]

    def get_historical_quotes(self, stock, interval, span, bounds=Bounds.REGULAR):
        self.endpoint_ = """Fetch historical data for stock.

        Note: valid interval/span configs
            interval = 5minute | 10minute + span = day, week
            interval = day + span = year
            interval = week
            TODO: NEEDS TESTS

        Args:
            stock (str): stock ticker
            interval (str): resolution of data
            span (str): length of data
            bounds (:obj:`Bounds`, optional): 'extended' or 'regular' trading hours

        Returns:
            (:obj:`dict`) values returned from `historicals` endpoint

        """
        if type(stock) is str:
            stock = [stock]

        if isinstance(bounds, str):  # recast to Enum
            bounds = Bounds(bounds)

        historicals = urls.HISTORICALS.with_query(
            [
                ("symbols", ",".join(stock).upper()),
                ("interval", interval),
                ("span", span),
                ("bounds", bounds.name.lower()),
            ]
        )

        return self.get(historicals)

    def get_news(self, stock):
        """Fetch news endpoint.

        Args:
            stock (str): stock ticker

        Returns:
            (:obj:`dict`) values returned from `news` endpoint

        """

        return self.get(urls.build_news(stock.upper()))

    def get_watchlists(self):
        """Fetch watchlists endpoint and queries for
        each instrumented result aka stock details returned from the watchlist

        Returns:
            (:obj:`dict`): values returned from `watchlists` and `instrument` endpoints
        """

        res = []
        watchlist = self.get(urls.WATCHLISTS)
        if watchlist and "results" in watchlist:
            data = self.get(watchlist["results"][0]["url"])
            for rec in data["results"]:
                res.append(self.get(rec["instrument"]))

        return res

    def print_quote(self, stock=""):  # pragma: no cover
        """Print quote information.

        Args:
            stock (str): ticker to fetch

        Returns:
            None

        """

        data = self.get_quote_list(stock, "symbol,last_trade_price")
        for item in data:
            quote_str = item[0] + ": $" + item[1]
            print(quote_str)

    def print_quotes(self, stocks):  # pragma: no cover
        """Print a collection of stocks.

        Args:
            stocks (:obj:`list`): list of stocks to pirnt

        Returns:
            None

        """

        if stocks is None:
            return

        for stock in stocks:
            self.print_quote(stock)

    def ask_price(self, stock=""):
        """Get asking price for a stock.

        Note:
            queries `quote` endpoint, dict wrapper

        Args:
            stock (str): stock ticker

        Returns:
            (float): ask price

        """

        return self.get_quote_list(stock, "ask_price")

    def ask_size(self, stock=""):
        """Get ask size for a stock.

        Note:
            queries `quote` endpoint, dict wrapper

        Args:
            stock (str): stock ticker

        Returns:
            (int): ask size

        """

        return self.get_quote_list(stock, "ask_size")

    def bid_price(self, stock=""):
        """Get bid price for a stock.

        Note:
            queries `quote` endpoint, dict wrapper

        Args:
            stock (str): stock ticker

        Returns:
            (float): bid price

        """

        return self.get_quote_list(stock, "bid_price")

    def bid_size(self, stock=""):
        """Get bid size for a stock.

        Note:
            queries `quote` endpoint, dict wrapper

        Args:
            stock (str): stock ticker

        Returns:
            (int): bid size

        """

        return self.get_quote_list(stock, "bid_size")

    def last_trade_price(self, stock=""):
        """Get last trade price for a stock

        Note:
            queries `quote` endpoint, dict wrapper

        Args:
            stock (str): stock ticker

        Returns:
            (float): last trade price

        """

        return self.get_quote_list(stock, "last_trade_price")

    def previous_close(self, stock=""):
        """Get previous closing price for a stock.

        Note:
            queries `quote` endpoint, dict wrapper

        Args:
            stock (str): stock ticker

        Returns:
            (float): previous closing price

        """

        return self.get_quote_list(stock, "previous_close")

    def previous_close_date(self, stock=""):
        """Get previous closing date for a stock.

        Note:
            queries `quote` endpoint, dict wrapper

        Args:
            stock (str): stock ticker

        Returns:
            (str): previous close date

        """

        return self.get_quote_list(stock, "previous_close_date")

    def adjusted_previous_close(self, stock=""):
        """Get adjusted previous closing price for a stock.

        Note:
            queries `quote` endpoint, dict wrapper

        Args:
            stock (str): stock ticker

        Returns:
            (float): adjusted previous closing price

        """

        return self.get_quote_list(stock, "adjusted_previous_close")

    def symbol(self, stock=""):
        """Get symbol for a stock.

        Note:
            queries `quote` endpoint, dict wrapper

        Args:
            stock (str): stock ticker

        Returns:
            (str): stock symbol

        """

        return self.get_quote_list(stock, "symbol")

    def last_updated_at(self, stock=""):
        """Get last update datetime.

        Note:
            queries `quote` endpoint, dict wrapper

        Args:
            stock (str): stock ticker

        Returns:
            (str): last update datetime
        """

        return self.get_quote_list(stock, "last_updated_at")

    def last_updated_at_datetime(self, stock=""):
        """Get last updated datetime.

        Note:
            queries `quote` endpoint, dict wrapper
            `self.last_updated_at` returns time as `str` in format:
            'YYYY-MM-ddTHH:mm:ss:000Z'

        Args:
            stock (str): stock ticker

        Returns:
            (datetime): last update datetime

        """

        # Will be in format: 'YYYY-MM-ddTHH:mm:ss:000Z'
        datetime_string = self.last_updated_at(stock)
        result = dateutil.parser.parse(datetime_string)

        return result

    def get_account(self):
        """Fetch account information.

        Returns:
            (:obj:`dict`): `accounts` endpoint payload

        """

        res = self.get(urls.ACCOUNTS)

        return res["results"][0]

    def get_url(self, url):
        """Flat wrapper for fetching URL directly/"""

        return self.get(url)

    def get_popularity(self, stock=""):
        """Get the number of robinhood users who own the given stock

        Args:
            stock (str): stock ticker

        Returns:
            (int): number of users who own the stock

        """
        stock_instrument = self.get_url(self.quote_data(stock)["instrument"])["id"]
        return self.get_url(urls.build_instruments(stock_instrument, "popularity"))[
            "num_open_positions"
        ]

    def get_tickers_by_tag(self, tag=None):
        """Get a list of instruments belonging to a tag

        Args: tag - Tags may include but are not limited to:
            * top-movers
            * etf
            * 100-most-popular
            * mutual-fund
            * finance
            * cap-weighted
            * investment-trust-or-fund

        Returns:
            (List): a list of Ticker strings

        """
        instrument_list = self.get_url(urls.build_tags(tag))["instruments"]
        return [self.get_url(instrument)["symbol"] for instrument in instrument_list]

    ###########################################################################
    #                           GET OPTIONS INFO                              #
    ###########################################################################

    def get_options(self, stock, expiration_dates, option_type):
        """Get a list (chain) of options contracts belonging to a particular stock

        Args: stock ticker (str), list of expiration dates to filter on
            (YYYY-MM-DD), and whether or not its a 'put' or a 'call' option type
            (str).

        Returns:
            Options Contracts (List): a list (chain) of contracts for a given \
            underlying equity instrument

        """
        instrument_id = self.get_url(self.quote_data(stock)["instrument"])["id"]
        if type(expiration_dates) == list:
            _expiration_dates_string = ",".join(expiration_dates)
        else:
            _expiration_dates_string = expiration_dates
        chain_id = self.get_url(urls.build_chain(instrument_id))["results"][0]["id"]
        return [
            contract
            for contract in self.get_url(
                urls.options(chain_id, _expiration_dates_string, option_type)
            )["results"]
        ]

    # TODO: the endpoint `option_market_data` doesn't exist
    # def get_option_market_data(self, optionid):
    #     """Gets a list of market data for a given optionid.
    #
    #     Args: (str) option id
    #
    #     Returns: dictionary of options market data.
    #
    #     """
    #     market_data = {}
    #     try:
    #         market_data = self.get_url(endpoints.option_market_data(optionid)) or {}
    #     except requests.exceptions.HTTPError:
    #         raise InvalidOptionId()
    #     return market_data

    def options_owned(self):
        options = self.get_url(urls.OPTIONS_BASE.join(URL("positions/?nonzero=true")))
        options = options["results"]
        return options

    def get_option_marketdata(self, option_id):
        info = self.get_url(urls.MARKET_DATA_BASE.join(URL(f"options/{option_id}/")))
        return info

    def get_option_chainid(self, symbol):
        stock_info = self.get_url(urls.INSTRUMENTS_BASE.with_query(symbol=symbol))
        instrument_id = stock_info["results"][0]["id"]
        url = urls.OPTIONS_BASE.join(URL("chains/"))
        chains = self.get_url(url.with_query(equity_instrument_ids=instrument_id))
        chains = chains["results"]
        chain_id = None

        for chain in chains:
            if chain["can_open_position"]:
                chain_id = chain["id"]

        return chain_id

    def get_option_quote(self, symbol, strike, expiry, otype, state="active"):
        url = urls.OPTIONS_BASE.join(URL("instruments/"))
        params = {
            "chain_symbol": symbol,
            "strike_price": strike,
            "expiration_dates": expiry,
            "type": otype,
            "state": state,
        }
        # symbol, strike, expiry, otype should uniquely define an option
        results = self.get_url(url.with_query(**params)).get("results")
        if not results:
            return
        else:
            option_id = results[0]["id"]
            result = self.get_option_marketdata(option_id)
            params["ask"] = "{} x {}".format(result["ask_size"], result["ask_price"])
            params["bid"] = "{} x {}".format(result["bid_size"], result["bid_price"])
            return params

    ###########################################################################
    #                           GET FUNDAMENTALS
    ###########################################################################

    def get_fundamentals(self, stock=""):
        """Find stock fundamentals data

        Args:
            (str): stock ticker

        Returns:
            (:obj:`dict`): contents of `fundamentals` endpoint

        """

        # Prompt for stock if not entered
        if not stock:  # pragma: no cover
            stock = input("Symbol: ")

        url = str(urls.build_fundamentals(str(stock.upper())))

        # Check for validity of symbol
        try:
            data = self.get(url)
        except requests.exceptions.HTTPError:
            raise InvalidTickerSymbol()

        return data

    def fundamentals(self, stock=""):
        """Wrapper for get_fundamentlals function"""

        return self.get_fundamentals(stock)

    ###########################################################################
    #                           PORTFOLIOS DATA
    ###########################################################################

    def portfolio(self):
        """Returns the user's portfolio data"""

        return self.get(urls.PORTFOLIOS, schema=PortfolioSchema())

    def order_history(self, orderId=None):
        """Wrapper for portfolios

        Optional Args: add an order ID to retrieve information about a single order.

        Returns:
            (:obj:`dict`): JSON dict from getting orders

        """

        return self.get(urls.build_orders(orderId))

    def dividends(self):
        """Wrapper for portfolios

        Returns:
            (:obj: `dict`): JSON dict from getting dividends

        """

        return self.get(urls.DIVIDENDS)

    ###########################################################################
    #                           POSITIONS DATA
    ###########################################################################

    def positions(self):
        """Returns the user's positions data

        Returns:
            (:object: `dict`): JSON dict from getting positions

        """

        return self.get(urls.POSITIONS)

    def securities_owned(self):
        """Returns list of securities' symbols that the user has shares in

        Returns:
            (:object: `dict`): Non-zero positions

        """

        return self.get(str(urls.POSITIONS) + "?nonzero=true")

    ###########################################################################
    #                               PLACE ORDER
    ###########################################################################

    def place_market_buy_order(
        self, instrument_URL=None, symbol=None, time_in_force=None, quantity=None
    ):
        """Wrapper for placing market buy orders

        Notes:
            If only one of the instrument_URL or symbol are passed as
            arguments the other will be looked up automatically.

        Args:
            instrument_URL (str): The RH URL of the instrument
            symbol (str): The ticker symbol of the instrument
            time_in_force (str): 'GFD' or 'GTC' (day or until cancelled)
            quantity (int): Number of shares to buy

        Returns:
            (:obj:`requests.request`): result from `orders` put command

        """
        return self.submit_buy_order(
            order_type="market",
            trigger="immediate",
            side="buy",
            instrument_URL=instrument_URL,
            symbol=symbol,
            time_in_force=time_in_force,
            quantity=quantity,
        )

    def place_limit_buy_order(
        self,
        instrument_URL=None,
        symbol=None,
        time_in_force=None,
        price=None,
        quantity=None,
    ):
        """Wrapper for placing limit buy orders

        Notes:
            If only one of the instrument_URL or symbol are passed as
            arguments the other will be looked up automatically.

        Args:
            instrument_URL (str): The RH URL of the instrument
            symbol (str): The ticker symbol of the instrument
            time_in_force (str): 'GFD' or 'GTC' (day or until cancelled)
            price (float): The max price you're willing to pay per share
            quantity (int): Number of shares to buy

        Returns:
            (:obj:`requests.request`): result from `orders` put command

        """
        return self.submit_buy_order(
            order_type="limit",
            trigger="immediate",
            side="buy",
            instrument_URL=instrument_URL,
            symbol=symbol,
            time_in_force=time_in_force,
            price=price,
            quantity=quantity,
        )

    def place_stop_loss_buy_order(
        self,
        instrument_URL=None,
        symbol=None,
        time_in_force=None,
        stop_price=None,
        quantity=None,
    ):
        """Wrapper for placing stop loss buy orders

        Notes:
            If only one of the instrument_URL or symbol are passed as
            arguments the other will be looked up automatically.

        Args:
            instrument_URL (str): The RH URL of the instrument
            symbol (str): The ticker symbol of the instrument
            time_in_force (str): 'GFD' or 'GTC' (day or until cancelled)
            stop_price (float): The price at which this becomes a market order
            quantity (int): Number of shares to buy

        Returns:
            (:obj:`requests.request`): result from `orders` put command

        """
        return self.submit_buy_order(
            order_type="market",
            trigger="stop",
            side="buy",
            instrument_URL=instrument_URL,
            symbol=symbol,
            time_in_force=time_in_force,
            stop_price=stop_price,
            quantity=quantity,
        )

    def place_stop_limit_buy_order(
        self,
        instrument_URL=None,
        symbol=None,
        time_in_force=None,
        stop_price=None,
        price=None,
        quantity=None,
    ):
        """Wrapper for placing stop limit buy orders

        Notes:
            If only one of the instrument_URL or symbol are passed as
            arguments the other will be looked up automatically.

        Args:
            instrument_URL (str): The RH URL of the instrument
            symbol (str): The ticker symbol of the instrument
            time_in_force (str): 'GFD' or 'GTC' (day or until cancelled)
            stop_price (float): The price at which this becomes a limit order
            price (float): The max price you're willing to pay per share
            quantity (int): Number of shares to buy

        Returns:
            (:obj:`requests.request`): result from `orders` put command

        """
        return self.submit_buy_order(
            order_type="limit",
            trigger="stop",
            side="buy",
            instrument_URL=instrument_URL,
            symbol=symbol,
            time_in_force=time_in_force,
            stop_price=stop_price,
            price=price,
            quantity=quantity,
        )

    def place_market_sell_order(
        self, instrument_URL=None, symbol=None, time_in_force=None, quantity=None
    ):
        """Wrapper for placing market sell orders

        Notes:
            If only one of the instrument_URL or symbol are passed as
            arguments the other will be looked up automatically.

        Args:
            instrument_URL (str): The RH URL of the instrument
            symbol (str): The ticker symbol of the instrument
            time_in_force (str): 'GFD' or 'GTC' (day or until cancelled)
            quantity (int): Number of shares to sell

        Returns:
            (:obj:`requests.request`): result from `orders` put command

        """
        return self.submit_sell_order(
            order_type="market",
            trigger="immediate",
            side="sell",
            instrument_URL=instrument_URL,
            symbol=symbol,
            time_in_force=time_in_force,
            quantity=quantity,
        )

    def place_limit_sell_order(
        self,
        instrument_URL=None,
        symbol=None,
        time_in_force=None,
        price=None,
        quantity=None,
    ):
        """Wrapper for placing limit sell orders

        Notes:
            If only one of the instrument_URL or symbol are passed as
            arguments the other will be looked up automatically.

        Args:
            instrument_URL (str): The RH URL of the instrument
            symbol (str): The ticker symbol of the instrument
            time_in_force (str): 'GFD' or 'GTC' (day or until cancelled)
            price (float): The minimum price you're willing to get per share
            quantity (int): Number of shares to sell

        Returns:
            (:obj:`requests.request`): result from `orders` put command

        """
        return self.submit_sell_order(
            order_type="limit",
            trigger="immediate",
            side="sell",
            instrument_URL=instrument_URL,
            symbol=symbol,
            time_in_force=time_in_force,
            price=price,
            quantity=quantity,
        )

    def place_stop_loss_sell_order(
        self,
        instrument_URL=None,
        symbol=None,
        time_in_force=None,
        stop_price=None,
        quantity=None,
    ):
        """Wrapper for placing stop loss sell orders

        Notes:
            If only one of the instrument_URL or symbol are passed as
            arguments the other will be looked up automatically.

        Args:
            instrument_URL (str): The RH URL of the instrument
            symbol (str): The ticker symbol of the instrument
            time_in_force (str): 'GFD' or 'GTC' (day or until cancelled)
            stop_price (float): The price at which this becomes a market order
            quantity (int): Number of shares to sell

        Returns:
            (:obj:`requests.request`): result from `orders` put command

        """
        return self.submit_sell_order(
            order_type="market",
            trigger="stop",
            side="sell",
            instrument_URL=instrument_URL,
            symbol=symbol,
            time_in_force=time_in_force,
            stop_price=stop_price,
            quantity=quantity,
        )

    def place_stop_limit_sell_order(
        self,
        instrument_URL=None,
        symbol=None,
        time_in_force=None,
        price=None,
        stop_price=None,
        quantity=None,
    ):
        """Wrapper for placing stop limit sell orders

        Notes:
            If only one of the instrument_URL or symbol are passed as
            arguments the other will be looked up automatically.

        Args:
            instrument_URL (str): The RH URL of the instrument
            symbol (str): The ticker symbol of the instrument
            time_in_force (str): 'GFD' or 'GTC' (day or until cancelled)
            stop_price (float): The price at which this becomes a limit order
            price (float): The max price you're willing to get per share
            quantity (int): Number of shares to sell

        Returns:
            (:obj:`requests.request`): result from `orders` put command

        """
        return self.submit_sell_order(
            order_type="limit",
            trigger="stop",
            side="sell",
            instrument_URL=instrument_URL,
            symbol=symbol,
            time_in_force=time_in_force,
            stop_price=stop_price,
            price=price,
            quantity=quantity,
        )

    # TODO: fix the function complexity
    def submit_sell_order(  # noqa: C901
        self,
        instrument_URL=None,
        symbol=None,
        order_type=None,
        time_in_force=None,
        trigger=None,
        price=None,
        stop_price=None,
        quantity=None,
        side=None,
    ):
        """Submits order to Robinhood

        Notes:
            This is normally not called directly.  Most programs should use
            one of the following instead:

                place_market_buy_order()
                place_limit_buy_order()
                place_stop_loss_buy_order()
                place_stop_limit_buy_order()
                place_market_sell_order()
                place_limit_sell_order()
                place_stop_loss_sell_order()
                place_stop_limit_sell_order()

        Args:
            instrument_URL (str): the RH URL for the instrument
            symbol (str): the ticker symbol for the instrument
            order_type (str): 'MARKET' or 'LIMIT'
            time_in_force (:obj:`TIME_IN_FORCE`): GFD or GTC (day or
                                                   until cancelled)
            trigger (str): IMMEDIATE or STOP enum
            price (float): The share price you'll accept
            stop_price (float): The price at which the order becomes a
                                market or limit order
            quantity (int): The number of shares to buy/sell
            side (str): BUY or sell

        Returns:
            (:obj:`requests.request`): result from `orders` put command

        """

        # Used for default price input
        # Price is required, so we use the current bid price if it is not specified
        current_quote = self.get_quote(symbol)
        if (current_quote["bid_price"] == 0) or (current_quote["bid_price"] is None):
            current_bid_price = current_quote["last_trade_price"]
        else:
            current_bid_price = current_quote["bid_price"]

        # Start with some parameter checks. I'm paranoid about $.
        if instrument_URL is None:
            if symbol is None:
                raise (
                    ValueError(
                        "Neither instrument_URL nor symbol were passed to "
                        "submit_sell_order"
                    )
                )
            raise (ValueError("Instrument_URL not passed to submit_sell_order"))

        if symbol is None:
            raise (ValueError("Symbol not passed to submit_sell_order"))

        if side is None:
            raise (
                ValueError("Order is neither buy nor sell in call to submit_sell_order")
            )

        if order_type is None:
            if price is None:
                if stop_price is None:
                    order_type = "market"
                else:
                    order_type = "limit"

        symbol = str(symbol).upper()
        order_type = str(order_type).lower()
        time_in_force = str(time_in_force).lower()
        trigger = str(trigger).lower()
        side = str(side).lower()

        if (order_type != "market") and (order_type != "limit"):
            raise (ValueError("Invalid order_type in call to submit_sell_order"))

        if order_type == "limit":
            if price is None:
                raise (
                    ValueError("Limit order has no price in call to submit_sell_order")
                )
            if price <= 0:
                raise (
                    ValueError(
                        "Price must be positive number in call to submit_sell_order"
                    )
                )

        if trigger == "stop":
            if stop_price is None:
                raise (
                    ValueError(
                        "Stop order has no stop_price in call to submit_sell_order"
                    )
                )
            if price <= 0:
                raise (
                    ValueError(
                        "Stop_price must be positive number in call to "
                        "submit_sell_order"
                    )
                )

        if stop_price is not None:
            if trigger != "stop":
                raise (
                    ValueError(
                        "Stop price set for non-stop order in call to submit_sell_order"
                    )
                )

        if price is None:
            if order_type == "limit":
                raise (
                    ValueError("Limit order has no price in call to submit_sell_order")
                )

        if price is not None:
            if order_type.lower() == "market":
                raise (
                    ValueError(
                        "Market order has price limit in call to submit_sell_order"
                    )
                )
            price = float(price)
        else:
            price = current_bid_price  # default to current bid price

        if quantity is None:
            raise (ValueError("No quantity specified in call to submit_sell_order"))

        quantity = int(quantity)

        if quantity <= 0:
            raise (
                ValueError(
                    "Quantity must be positive number in call to submit_sell_order"
                )
            )

        payload = {}

        for field, value in [
            ("account", self.get_account()["url"]),
            ("instrument", instrument_URL),
            ("symbol", symbol),
            ("type", order_type),
            ("time_in_force", time_in_force),
            ("trigger", trigger),
            ("price", price),
            ("stop_price", stop_price),
            ("quantity", quantity),
            ("side", side),
        ]:
            if value is not None:
                payload[field] = value

        res = self.post(urls.build_orders(), data=payload)
        return res

    # TODO: Fix function complexity
    def submit_buy_order(  # noqa: C901
        self,
        instrument_URL=None,
        symbol=None,
        order_type=None,
        time_in_force=None,
        trigger=None,
        price=None,
        stop_price=None,
        quantity=None,
        side=None,
    ):
        """Submits buy order to Robinhood

        Notes:
            This is normally not called directly.  Most programs should use
            one of the following instead:

                place_market_buy_order()
                place_limit_buy_order()
                place_stop_loss_buy_order()
                place_stop_limit_buy_order()
                place_market_sell_order()
                place_limit_sell_order()
                place_stop_loss_sell_order()
                place_stop_limit_sell_order()

        Args:
            instrument_URL (str): the RH URL for the instrument
            symbol (str): the ticker symbol for the instrument
            order_type (str): 'market' or 'limit'
            time_in_force (:obj:`TIME_IN_FORCE`): 'gfd' or 'gtc' (day or
                                                   until cancelled)
            trigger (str): 'immediate' or 'stop' enum
            price (float): The share price you'll accept
            stop_price (float): The price at which the order becomes a
                                market or limit order
            quantity (int): The number of shares to buy/sell
            side (str): BUY or sell

        Returns:
            (:obj:`requests.request`): result from `orders` put command

        """

        # Used for default price input
        # Price is required, so we use the current ask price if it is not specified
        current_quote = self.get_quote(symbol)
        if (current_quote["ask_price"] == 0) or (current_quote["ask_price"] is None):
            current_ask_price = current_quote["last_trade_price"]
        else:
            current_ask_price = current_quote["ask_price"]

        # Start with some parameter checks. I'm paranoid about $.
        if instrument_URL is None:
            if symbol is None:
                raise (
                    ValueError(
                        "Neither instrument_URL nor symbol were passed to "
                        "submit_buy_order"
                    )
                )
            raise (ValueError("Instrument_URL not passed to submit_buy_order"))

        if symbol is None:
            raise (ValueError("Symbol not passed to submit_buy_order"))

        if side is None:
            raise (
                ValueError("Order is neither buy nor sell in call to submit_buy_order")
            )

        if order_type is None:
            if price is None:
                if stop_price is None:
                    order_type = "market"
                else:
                    order_type = "limit"

        symbol = str(symbol).upper()
        order_type = str(order_type).lower()
        time_in_force = str(time_in_force).lower()
        trigger = str(trigger).lower()
        side = str(side).lower()

        if (order_type != "market") and (order_type != "limit"):
            raise (ValueError("Invalid order_type in call to submit_buy_order"))

        if order_type == "limit":
            if price is None:
                raise (
                    ValueError("Limit order has no price in call to submit_buy_order")
                )
            if price <= 0:
                raise (
                    ValueError(
                        "Price must be positive number in call to submit_buy_order"
                    )
                )

        if trigger == "stop":
            if stop_price is None:
                raise (
                    ValueError(
                        "Stop order has no stop_price in call to submit_buy_order"
                    )
                )
            if price <= 0:
                raise (
                    ValueError(
                        "Stop_price must be positive number in call to submit_buy_order"
                    )
                )

        if stop_price is not None:
            if trigger != "stop":
                raise (
                    ValueError(
                        "Stop price set for non-stop order in call to submit_buy_order"
                    )
                )

        if price is None:
            if order_type == "limit":
                raise (
                    ValueError("Limit order has no price in call to submit_buy_order")
                )

        if price is not None:
            if order_type.lower() == "market":
                raise (
                    ValueError(
                        "Market order has price limit in call to submit_buy_order"
                    )
                )
            price = float(price)
        else:
            price = current_ask_price  # default to current ask price

        if quantity is None:
            raise (ValueError("No quantity specified in call to submit_buy_order"))

        quantity = int(quantity)

        if quantity <= 0:
            raise (
                ValueError(
                    "Quantity must be positive number in call to submit_buy_order"
                )
            )

        payload = {}

        for field, value in [
            ("account", self.get_account()["url"]),
            ("instrument", instrument_URL),
            ("symbol", symbol),
            ("type", order_type),
            ("time_in_force", time_in_force),
            ("trigger", trigger),
            ("price", price),
            ("stop_price", stop_price),
            ("quantity", quantity),
            ("side", side),
        ]:
            if value is not None:
                payload[field] = value

        res = self.post(urls.build_orders(), data=payload)
        return res

    def place_order(
        self,
        instrument,
        quantity=1,
        price=0.0,
        transaction=None,
        trigger="immediate",
        order="market",
        time_in_force="gfd",
    ):
        """Place an order with Robinhood

        Args:
            instrument (dict): the RH URL and symbol in dict for the instrument to
                be traded
            quantity (int): quantity of stocks in order
            bid_price (float): price for order
            transaction (:obj:`Transaction`): BUY or SELL enum
            trigger (:obj:`Trigger`): IMMEDIATE or STOP enum
            order (:obj:`Order`): MARKET or LIMIT
            time_in_force (:obj:`TIME_IN_FORCE`): GFD or GTC (day or until
                cancelled)

        Returns:
            (:obj:`requests.request`): result from `orders` put command

        """

        if isinstance(transaction, str):
            transaction = Transaction(transaction)

        if not price:
            price = self.quote_data(instrument["symbol"])["bid_price"]

            if (price == 0) or (price is None):
                price = self.quote_data(instrument["symbol"])["last_trade_price"]

        payload = {
            "account": self.get_account()["url"],
            "instrument": unquote(instrument["url"]),
            "symbol": instrument["symbol"],
            "type": order.lower(),
            "time_in_force": time_in_force.lower(),
            "trigger": trigger,
            "quantity": quantity,
            "side": transaction.name.lower(),
        }

        if order.lower() == "stop":
            payload["stop_price"] = float(price)
        else:
            payload["price"] = float(price)

        res = self.post(urls.build_orders(), data=payload)
        return res

    def place_buy_order(self, instrument, quantity, ask_price=0.0):
        """Wrapper for placing buy orders

        Args:
            instrument (dict): the RH URL and symbol in dict for the instrument to
                be traded
            quantity (int): quantity of stocks in order
            ask_price (float): price for order (OPTIONAL! If not given, ask_price is
                automatic.)

        Returns:
            (:obj:`requests.request`): result from `orders` put command

        """

        if not ask_price:
            ask_price = self.quote_data(instrument["symbol"])["ask_price"]

            if (ask_price == 0) or (ask_price is None):
                ask_price = self.quote_data(instrument["symbol"])["last_trade_price"]

        transaction = Transaction.BUY

        return self.place_order(instrument, quantity, ask_price, transaction)

    def place_sell_order(self, instrument, quantity, bid_price=0.0):
        """Wrapper for placing sell orders

        Args:
            instrument (dict): the RH URL and symbol in dict for the instrument to
                be traded
            quantity (int): quantity of stocks in order
            bid_price (float): price for order (OPTIONAL! If not given, bid_price is
                automatic.)

        Returns:
            (:obj:`requests.request`): result from `orders` put command

        """
        if not bid_price:
            bid_price = self.quote_data(instrument["symbol"])["bid_price"]

            if (bid_price == 0) or (bid_price is None):
                bid_price = self.quote_data(instrument["symbol"])["last_trade_price"]

        transaction = Transaction.SELL

        return self.place_order(instrument, quantity, bid_price, transaction)

    ##############################
    # GET OPEN ORDER(S)
    ##############################

    def get_open_orders(self):
        """Returns all currently open (cancellable) orders.

        If not orders are currently open, `None` is returned.

        TODO: Is there a way to get these from the API endpoint without stepping through
            order history?
        """

        open_orders = []
        orders = self.order_history()
        for order in orders["results"]:
            if order["cancel"] is not None:
                open_orders.append(order)

        return open_orders

    ##############################
    #        CANCEL ORDER        #
    ##############################

    # TODO: Fix function complexity
    def cancel_order(self, order_id):  # noqa: C901
        """Cancels specified order and returns the response.

        If order cannot be cancelled, `None` is returned.
        (results from `orders` command).

        Args:
            order_id (str or dict): Order ID string that is to be cancelled or open
                order dict returned from order get.

        Returns:
            (:obj:`requests.request`): result from `orders` put command

        """
        if isinstance(order_id, str):
            try:
                order = self.get(urls.build_orders(order_id))
            except (requests.exceptions.HTTPError) as err_msg:
                raise ValueError(
                    "Failed to get Order for ID: "
                    + order_id
                    + "\n Error message: "
                    + repr(err_msg)
                )

            if order.get("cancel") is not None:
                try:
                    res = self.post(order["cancel"])
                    return res
                except requests.exceptions.HTTPError:
                    try:
                        # sometimes Robinhood asks for another log in when placing an
                        # order
                        res = self.post(order["cancel"])
                        return res
                    except (requests.exceptions.HTTPError) as err_msg:
                        raise ValueError(
                            "Failed to cancel order ID: "
                            + order_id
                            + "\n Error message: "
                            + repr(err_msg)
                        )
                        return None

        elif isinstance(order_id, dict):
            order_id = order_id["id"]
            try:
                order = self.get(urls.build_orders(order_id))
            except (requests.exceptions.HTTPError) as err_msg:
                raise ValueError(
                    "Failed to get Order for ID: "
                    + order_id
                    + "\n Error message: "
                    + repr(err_msg)
                )

            if order.get("cancel") is not None:
                try:
                    res = self.post(order["cancel"])
                    return res
                except requests.exceptions.HTTPError:
                    try:
                        # sometimes Robinhood asks for another log in when placing an
                        # order
                        res = self.post(order["cancel"])
                        return res
                    except requests.exceptions.HTTPError as err_msg:
                        raise ValueError(
                            "Failed to cancel order ID: "
                            + order_id
                            + "\n Error message: "
                            + repr(err_msg)
                        )
                        return None

        elif not isinstance(order_id, str) or not isinstance(order_id, dict):
            raise ValueError(
                "Cancelling orders requires a valid order_id string or open order"
                "dictionary"
            )

        # Order type cannot be cancelled without a valid cancel link
        else:
            raise ValueError("Unable to cancel order ID: " + order_id)


class RobinhoodSchema(SessionManagerSchema):
    """Schema for the Robinhood class."""

    __model__ = Robinhood
